
from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of reading & writing medical files. ",
    help = "Use `NII` and `DCM` to cope with files.",
    requires = ["nibabel", "pydicom"]
)

__all__ = """
    IMG
    dcm2nii
    nii2dcm
    affine2orient
    affine2spacing
    dim_of_orient_axis
""".split()

import os
from datetime import datetime as dt

with __info__:
    import SimpleITK as sitk
    import batorch as bt
    import numpy as np
    from pycamia import Path, Workflow, SPrint, Error
    from pycamia import no_print, touch, avouch, alias, get_environ_vars
    from pycamia import to_tuple, to_list, get_alphas, get_digits, arg_tuple
    from pycamia import item, argmax, sublist, map_ele
    from pycamia.numpyop import toI
    from pyoverload import overload, callable
    from pyoverload import to_torch_dtype, to_numpy_type, to_dtype

available_modalities = ['PNG'] # TODO: MHA
sitk_supported_modalities = ['NII', 'NRRD']
sitk_default_exts = {'NII': '.nii.gz', 'NRRD': '.nrrd'}
sitk_modality_exts = {'.nii': 'NII', '.nii.gz': 'NII', '.nrrd': 'NRRD'}
available_modalities.extend(sitk_supported_modalities)

try: import nibabel as nib; import nrrd # pip install pynrrd
except ModuleNotFoundError: pass
try: import pydicom as dcm; available_modalities.append('DCM')
except [ImportError, ModuleNotFoundError]: pass

run = Workflow(*available_modalities, verbose=False)
cdate = lambda: dt.now().strftime("%Y%m%d")
ctime = lambda: dt.now().strftime("%H%M%S")
cmilitime = lambda: dt.now().strftime("%H%M%S.%f")[:-3]
cstamp = lambda: dt.now().strftime("%Y%m%d%H%M%S")
cmilistamp = lambda: dt.now().strftime("%Y%m%d%H%M%S%f")[:-2]
get_uid7 = lambda: "1.2.100.100000.100000.111." + cstamp()
get_uid10 = lambda: "1.3.10.100000.11.10000.1.0.1000." + cmilistamp()

basic_tags = {
    "0010|0010": "Unknown", # Patient Name
    "0010|0020": "00000001", # Patient ID
    "0010|0030": "20000101", # Patient Birth Date
    "0010|0040": 'M', # Patient Sex
    "0010|1010": "020Y", # Patient Age
    "0010|1020": "100", # Patient Height
    "0010|1030": "200", # Patient Weight
    "0008|0020": cdate(), # Study Date
    "0008|0030": ctime(), # Study Time
    "0008|0050": cdate() + "00010001", # Accession Number
    "0008|0060": "Unknown",  # Modality
    "0020|000D": get_uid7(), # Study Instance UID, for machine consumption
    "0020|0010": "111111111", # Study ID, for human consumption
    "0020|0052": get_uid10(), # Frame of Reference UID
    # "0028|1050": '197.64', # Window Center
    # "0028|1051": '375.23', # Window Width
}

with run("DCM", "NII"), run.all_tags:
    def dcm2nii(file1, file2):
        IMG(file1).astype(IMG.nii).save(file2)

    def nii2dcm(file1, file2):
        IMG(file1).astype(IMG.dcm).save(file2)
    
@bt.batorch_wrapper
def affine2orient(affine: bt.Tensor):
    if affine.n_dim > 2:
        affine = affine.flatten(0, -3)
        n_batch = affine.size(0)
        ret_list = True
    else: affine = affine.unsqueeze(0); n_batch = 1; ret_list = False
    orients = []
    for ia in range(n_batch):
        aff_abs = bt.abs(affine[ia, :-1, :-1])
        n_dim = aff_abs.shape[-1]
        orient = [''] * n_dim
        for d in range(n_dim):
            i, j = aff_abs.argmax().indices
            orient[j] = ['RL', 'AP', 'SI'][i][affine[ia, i, j] > 0]
            aff_abs[i] = 0
            aff_abs[:, j] = 0
        orients.append(''.join(orient))
    if not ret_list: return orients[0]
    else: return orients
    
@bt.batorch_wrapper
def affine2spacing(affine: bt.Tensor):
    if affine.n_dim > 2:
        affine = affine.flatten(0, -3)
        n_batch = affine.size(0)
        return [tuple(bt.sqrt(((affine[ia, :-1, :-1]) ** 2).sum(-2)).round(decimals=4).tolist()) for ia in range(n_batch)]
    else: return tuple(bt.sqrt(((affine[:-1, :-1]) ** 2).sum(-2)).round(decimals=4).tolist())
    
def dim_of_orient_axis(orient, axis):
    dim = None
    for d in axis:
        if d in orient: dim = orient.index(d)
    if dim is None: raise TypeError(f"Invalid orient axis {axis} for orientation {orient}")
    return dim
        
class Spacing(tuple): ...

class IMG:
    
    with run("PNG"), run.use_tag: png = PNG = 'PNG'
    with run("DCM"), run.use_tag: dcm = DCM = 'DCM'
    with run(sitk_supported_modalities), run.all_tags:
        for m in sitk_supported_modalities:
            exec(f"{m.lower()} = {m.upper()} = '{m.upper()}'")
    
    basic_info = "SeriesNumber SeriesDescription SeriesTime Shape".split()

    @overload
    def __init__(self, path: str, **kwargs):
        if path.__class__ != Path: path = Path(path)
        self.unresolved_warning = False
        self.component_seperated = False
        self.mismatch_seperated = False
        done = False
        if path.is_file():
            with run(sitk_supported_modalities), run.any_tag:
                if any([path | e for e in sitk_modality_exts.keys()]): self.__init_sitk__(path, **kwargs); done =True
            with run("DCM"), run.use_tag:
                if path | 'dcm' or path | 'ima': self.__init_dcm__(path.parent, **kwargs); done =True
        elif path.is_dir():
            with run("DCM"), run.use_tag:
                if len([x for x in path if x | 'dcm' or x | 'ima']) > 0: self.__init_dcm__(path, **kwargs); done =True
        if not done:
            if not path.exists(): raise FileNotFoundError(f"Cannot find file/folder {path}. ")
            self.ftype = kwargs.get('astype', None)
            if self.ftype is None: raise TypeError(f"micomputing.IMG failed to recognize type of {path}. Use `astype=IMG.nii` or other image format to identify it. ")
            with run(sitk_supported_modalities), run.any_tag:
                if self.ftype in sitk_supported_modalities: self.__init_sitk__(path, **kwargs); done =True
            with run("DCM"), run.use_tag:
                if self.ftype in ("DCM", "IMA"): self.__init_dcm__(path, **kwargs); done =True
        if len(self.basics) <= 0: self.basics = basic_tags.copy()

    @overload
    def __init__(self, x: sitk.Image, **kwargs):
        self.unresolved_warning = False
        self.has_series = False
        self.ftype = kwargs.get('ftype', None)
        self.sid = kwargs.get('sid', get_uid10())
        self.image = x
        self.basics = kwargs.get('basics', basic_tags.copy())
        self.file_list = kwargs.get('file_list', [])
        self.info = kwargs.get('info', {})
        if 'Shape' not in self.info:
            if x.GetDepth(): self.info['Shape'] = x.GetSize() + (x.GetDepth(),)
            else: self.info['Shape'] = x.GetSize()

    @overload
    def __init__(self, x: bt.Tensor, **kwargs):
        x = x.squeeze()
        if not x.dtype.is_floating_point: x = x.type(bt.int16)
        if x.n_space_dim >= 3:
            x = bt.permute_space(x, 2, 1, 0, *range(3, x.n_space_dim))
            if x.n_space_dim > 3: x.with_sz_sequence_dim(3 - x.n_space_dim)
        else: x = x.transpose(0, 1)
        x = x.detach().cpu()
        self.unresolved_warning = False
        self.has_series = x.has_special
        self.ftype = kwargs.get('ftype', None)
        self.basics = kwargs.get('basics', basic_tags)
        if self.has_series:
            if x.has_batch and x.has_channel: x = x.standard_shape().flatten(0, 1).with_batch_dimension(0)
            if x.has_batch: series = x.split(1, {})
            else: series = x.split(1, [])
            self.images = {}
            self.file_lists = {}
            self.series_infos = {}
            self.series_IDs = []
            for i, s in enumerate(series):
                s = s.squeeze()
                sid = get_uid10()
                self.series_IDs.append(sid)
                self.images[sid] = sitk.GetImageFromArray(s)
                self.file_lists[sid] = []
                self.series_infos[sid] = dict(
                    SeriesNumber = str(i + 1), 
                    SeriesDescription = f"Series {i + 1}", 
                    Shape = s.shape
                )
            return
        self.sid = kwargs.get('sid', get_uid10())
        self.image = sitk.GetImageFromArray(x)
        self.file_list = kwargs.get('file_list', [])
        self.info = kwargs.get('info', {})
        if 'Shape' not in self.info:
            self.info['Shape'] = x.space

    with run(sitk_supported_modalities), run.any_tag:
        def __init_sitk__(self, path, **kwargs):
            self.has_series = False
            self.ftype = item([t for e, t in sitk_modality_exts.items() if path | e])
            self.sid = ''
            self.image = sitk.ReadImage(path)
            self.file_list = [path]
            self.basics = kwargs.get('basics', {})
            self.info = kwargs.get('info', {})
            self.info['Shape'] = self.image.GetSize()
    
    with run("DCM"), run.use_tag:       
        def __init_dcm__(self, path, **kwargs):
            """
            Initialize a dicom data structure

            Args:
                path (str): the path to the DICOM directory. 
                is_valid_series_ID (callable): a function that accepts a series id and returns whether it is valid.
                    Use `lambda x: x == a` to select one series and `lambda x: x != a` to exclude one. 
                is_valid_series (callable): a function that accepts a dictionary with keywords 'SeriesNumber', 'SeriesDescription', 
                    and 'Shape'. One should use them to identify whether the series ought to be preserved. 

            Raises:
                TypeError: Invalid DICOM files. 
            """
            self.has_series = True
            self.ftype = 'DCM'
            self.images = {}
            self.file_lists = {}
            self.series_infos = {}
            self.basics = {}
            self.series_IDs = sitk.ImageSeriesReader.GetGDCMSeriesIDs(path)
            if len(self.series_IDs) <= 0: raise TypeError(f"Not dicom directory at {path}.")
            is_valid_series_ID = kwargs.get('is_valid_series_ID', lambda x: True)
            if isinstance(is_valid_series_ID, str): is_valid_series_ID = lambda x: x == is_valid_series_ID
            if isinstance(is_valid_series_ID, list): is_valid_series_ID = lambda x: x in is_valid_series_ID
            self.series_IDs = [i for i in self.series_IDs if is_valid_series_ID(i)]
            if len(self.series_IDs) > 1:
                is_valid_series = kwargs.get('is_valid_series', lambda x: True)
                if isinstance(is_valid_series, str): is_valid_series = lambda x: is_valid_series in x.values()
                if isinstance(is_valid_series, list): is_valid_series = lambda x: any([i in is_valid_series for i in x.values()])
            else: is_valid_series = lambda x: True
            ids_to_delete = []
            for sid in self.series_IDs.copy():
                file_list = list(sitk.ImageSeriesReader.GetGDCMSeriesFileNames(path, seriesID = sid))
                with dcm.dcmread(file_list[0], stop_before_pixels=True, force=True) as dcm_header:
                    if len(self.basics) <= 0: self.basics = {t: touch(lambda: str(dcm_header[tuple(eval('0x'+u) for u in t.split('|'))].value), "") for t in basic_tags}
                    info = dict(SeriesNumber = dcm_header.SeriesNumber, SeriesDescription = dcm_header.SeriesDescription, SeriesTime = getattr(dcm_header, 'SeriesTime', '[N/A]'), Shape = (dcm_header.Rows, dcm_header.Columns, len(file_list)))
                    if not is_valid_series(info): ids_to_delete.append(sid); continue
                seperated = self._read_dcm(sid, file_list, info)
                if seperated: ids_to_delete.append(sid)
            for sid in ids_to_delete:
                self.series_IDs.remove(sid)
            self.series_IDs.sort(key=lambda x: tuple(map(eval, x.split('.'))))
            if len(self.series_IDs) == 1:
                self.has_series = False
                self.sid = self.series_IDs[0]
                self.image = self.images[self.sid]
                self.file_list = self.file_lists[self.sid]
                self.info = self.series_infos[self.sid]
                
        def _handle_dcm_size_exception(self, file_list, *_):
            sizes = []
            files = []
            for file_path in file_list:
                with dcm.dcmread(file_path, stop_before_pixels=True, force=True) as dcm_header:
                    size = (dcm_header.Rows, dcm_header.Columns)
                    if size in sizes: files[sizes.index(size)].append(file_path)
                    else: sizes.append(size); files.append([file_path])
            self.mismatch_seperated = True
            return [({'Shape': s + (len(f),)}, f) for s, f in zip(sizes, files)]
        
        def _handle_dcm_non_uniform_exception(self, file_list, series_name):
            all_med_params = {}
            med_param_keys = []
            med_param_names = []
            med_param_values = []
            for file_path in file_list:
                with dcm.dcmread(file_path, stop_before_pixels=True, force=True) as dcm_header:
                    if len(med_param_keys) == 0:
                        for k in list(dcm_header.keys()):
                            if '0018,' not in str(k) and str(k) != "(0008, 0008)": continue
                            med_param_keys.append(k)
                            med_param_names.append(dcm_header[k].name)
                            med_param_values.append(set())
                            
                    all_med_params.setdefault(file_path, [])
                    for i, k in enumerate(med_param_keys):
                        v = dcm_header[k].value
                        if not isinstance(v, (str, int, float, tuple)): v = str(v)
                        all_med_params[file_path].append(v)
                        med_param_values[i].add(v)
            critical_pos = [i for i, x in enumerate(med_param_values) if len(x) > 1]
            if len(critical_pos) == len(med_param_keys):
                file_list.sort()
                print(f"Non uniform sampling in {series_name}")
                print(f"Files: {file_list[0]}--{os.path.basename(file_list[-1])}")
                print("Failed to seperate: same medical parameters ")
                return
            med_names = []
            new_critical_pos = []
            for i in critical_pos:
                name = med_param_names[i]
                if 'time' in name.lower(): continue
                if 'name' in name.lower():
                    new_critical_pos = [i]
                    med_names = [name]
                    break
                new_critical_pos.append(i)
                med_names.append(name)
            if len(new_critical_pos) == 0: return [({}, file_list)]
            partition = {}
            for file_path in file_list:
                value_set = tuple(sublist(all_med_params[file_path], new_critical_pos))
                value_set = to_tuple(value_set)
                partition.setdefault(value_set, [])
                partition[value_set].append(file_path)
            value_sets = sorted(list(partition.keys()))
            return [(dict(zip(med_names, vs)), partition[vs]) for vs in value_sets]
                    
        def _read_dcm(self, sid, file_list, info):
            series_name = f"[{info['SeriesTime']}] Series {info['SeriesDescription']}<{info['SeriesNumber']}> {info['Shape']}"
            with dcm.dcmread(file_list[1 if len(file_list) > 1 else 0], stop_before_pixels=True, force=True) as dcm_header: pS = getattr(dcm_header, 'ImagePositionPatient', [0])[-1]
            with dcm.dcmread(file_list[-2 if len(file_list) > 1 else -1], stop_before_pixels=True, force=True) as dcm_header: pE = getattr(dcm_header, 'ImagePositionPatient', [0])[-1]
            if pS > pE: file_list = file_list[::-1]
            with no_print as rec:
                try:
                    self.series_infos[sid] = info
                    self.file_lists[sid] = file_list
                    series_reader = sitk.ImageSeriesReader()
                    series_reader.SetFileNames(file_list)
                    self.images[sid] = series_reader.Execute()
                    res = "DONE"
                except RuntimeError as e:
                    error = str(e)
                    if 'Size mismatch' in error or 'requested regionRequested' in error:
                        res = self._handle_dcm_size_exception(file_list, series_name)
                    else:
                        file_list.sort()
                        self.unresolved_warning = True
                        raise RuntimeError(f"Failure in {series_name} Files: {file_list[0]}--{os.path.basename(file_list[-1])}\n" + error)
            if "WARNING" in rec.string():
                if "Non uniform sampling" in str(rec):
                    res = self._handle_dcm_non_uniform_exception(file_list, series_name)
                else:
                    file_list.sort()
                    print(rec)
                    print(f"in {series_name} Files: {file_list[0]}--{os.path.basename(file_list[-1])}")
                    self.unresolved_warning = True; return False
            if res is None or isinstance(res, list) and len(res) == 0:
                self.unresolved_warning = True; return False
            if isinstance(res, str): return False
            if len(res) == 1: return False
            self.series_infos.pop(sid)
            self.file_lists.pop(sid)
            if sid in self.images: self.images.pop(sid)
            for i, (extra_info, files) in enumerate(res):
                key = sid + f'.{i+1}.0.0'
                new_info = info.copy()
                new_info.update(extra_info)
                new_info['SeriesDescription'] = '_'.join([new_info.get('SeriesDescription', '')] + sum([k.split() + [str(v)] for k, v in extra_info.items()], []))
                if 'Shape' in extra_info:
                    d = extra_info['Shape'][-1]
                    if d <= 3: continue
                    new_info['Depth'] = d
                sep = self._read_dcm(key, files, new_info)
                if not sep: self.series_IDs.append(key)
            return True

    @alias("__array__")
    def to_numpy(self, sid=None):
        """
        Get the array data item of IMG.
        Note: It is transposed for 3D arrays as the direct array by sitk is in size (n_z, n_x, n_y)
        """
        if self.has_series:
            if sid is None: raise TypeError("Converting micomputing.IMG with series to numpy. Please Identify Series ID. ")
            image = self.images[sid]
        else: image = self.image
        image_data = np.array(sitk.GetArrayFromImage(image))
        if image_data.ndim >= 3: return image_data.transpose(2, 1, 0, *range(3, image_data.ndim))
        else: return image_data.transpose(0, 1)

    @alias("__tensor__")
    def to_tensor(self, sid=None):
        """
        Get the tensor item of IMG.
        Note: It is transposed for 3D arrays as the direct array by sitk is in size (n_z, n_x, n_y)
        """
        data_array = self.to_numpy(sid=sid)
        tensor_dtype = to_torch_dtype(to_dtype(data_array.dtype))
        data_tensor = bt.tensor(data_array.astype(to_numpy_type(to_dtype(tensor_dtype))), dtype=tensor_dtype)
        data_tensor.image = self
        return data_tensor

    @bt.batorch_wrapper
    def from_array(self, array: bt.Tensor):
        array = array.squeeze()
        if not array.dtype.is_floating_point: array = array.type(bt.int16)
        if array.n_space_dim >= 3:
            array = bt.permute_space(array, 2, 1, 0, *range(3, array.n_space_dim))
            if array.n_space_dim > 3: array.with_sz_sequence_dim(3 - array.n_space_dim)
        else: array = array.transpose(0, 1)
        array = array.detach().cpu()
        old_image = self.image
        self.image = sitk.GetImageFromArray(array)
        self.image.SetDirection(old_image.GetDirection())
        self.image.SetSpacing(old_image.GetSpacing())
        self.image.SetOrigin(old_image.GetOrigin())
        return self
        
    def __str__(self):
        str_print = SPrint()
        if self.has_series:
            str_print(f"<micomputing.IMG with {len(self.series_IDs)} series, ftype={repr(self.ftype)}, dtype={repr(self.images[self.series_IDs[0]].GetPixelIDTypeAsString())}>")
            for sid in self.series_IDs:
                info = self.series_infos[sid]
                extra = info.copy()
                for k in IMG.basic_info: extra.pop(k)
                if len(extra) > 0: str_print(f"\t[{info['SeriesTime']}] Series {sid}: {info['SeriesDescription']}<{info['SeriesNumber']}>, Size: {info['Shape']}, {str(extra)}.")
                else: str_print(f"\t[{info['SeriesTime']}] Series {sid}: {info['SeriesDescription']}<{info['SeriesNumber']}>, Size: {info['Shape']}.")
        else: str_print(f"<micomputing.IMG object, Size: {self.info['Shape']}, ftype={repr(self.ftype)}, dtype={repr(self.image.GetPixelIDTypeAsString())}>")
        return str_print.text
        
    def split(self):
        if not self.has_series: raise TypeError("micomputing.IMG without series is not iterable. ")
        collection = []
        for sid in self.series_IDs:
            collection.append(self[sid])
        return collection
    
    def __getitem__(self, sid):
        if not self.has_series: raise TypeError("micomputing.IMG without series is not iterable. ")
        if not isinstance(sid, str): sid = str(sid)
        if '.' not in sid:
            is_series_num = touch(lambda: int(sid), False)
            if is_series_num: sid = item([x for x in self.series_IDs if self.series_infos[x]['SeriesNumber'] == sid])
            else: sid = item([x for x in self.series_IDs if self.series_infos[x]['SeriesDescription'] == sid])
        return IMG(self.images[sid], sid=sid, info=self.series_infos[sid], file_list=self.file_lists[sid], ftype=self.ftype)
    
    def __iter__(self):
        if not self.has_series: raise TypeError("micomputing.IMG without series is not iterable. ")
        for sid in self.series_IDs:
            yield self[sid]

    def __enter__(self): return self

    def __exit__(self, *args): return False

    def astype(self, ftype): self.ftype = ftype; return self
    
    def copy(self):
        if self.has_series:
            tmp_image = self.images[self.series_IDs[0]]
            output = IMG(tmp_image, ftype=self.ftype, basics=self.basics, file_list={}, info={})
            output.unresolved_warning = False
            output.has_series = True
            output.images = self.images.copy()
            output.file_lists = self.file_lists.copy()
            output.series_infos = self.series_infos.copy()
            output.basics = self.basics.copy()
            output.series_IDs = self.series_IDs.copy()
            return output
        return IMG(self.image, ftype=self.ftype, sid=self.sid, basics=self.basics, file_list=self.file_list, info=self.info)
    
    @property
    def path(self):
        if len(self.file_list) == 1: return self.file_list[0]
        return self.file_list[0].dirname
    
    @property
    @alias("affine_of")
    def affine(self, image=None):
        """
        Return the Niftii affine matrix that converts image indices to world coordinates. 
        Note: sitk contains direction in DICOM format with +x means left and +y means posterior.
        Hence, Niftii format of IMG requires the first two rows of the affine matrix being inversed. 
        """
        if image is None:
            if self.has_series: raise Error('Property')("No property 'affine' for micomputing.IMG with series. ")
            image = self.image
        if self.has_series and isinstance(image, str): image = self.images[image]
        dir = np.array(image.GetDirection())
        ndim = image.GetDimension()
        avouch(ndim ** 2 == len(dir))
        A = dir.reshape((ndim, ndim)) @ np.diag(np.array(image.GetSpacing()))
        b = np.array(image.GetOrigin()).reshape((ndim, 1))
        c = np.concatenate((np.zeros((1, ndim)), np.ones((1, 1))), 1)
        return np.diag([-1, -1] + [1] * (ndim - 1)) @ np.concatenate((np.concatenate((A, b), 1), c), 0)

    @affine.setter
    @alias("set_affine")
    def affine(self, aff, spacing=None, spacing2one=False):
        if self.has_series: raise Error('Property')("Cannot set property 'affine' for micomputing.IMG with series. ")
        ndim = aff.shape[-1] - 1
        try: aff = aff.reshape((ndim + 1, ndim + 1))
        except ValueError: raise Error('Value')(f"Cannot set affine matrix to image: expected shape ({ndim + 1}, {ndim + 1}) but got {aff.shape}")
        flipped_aff = np.diag([-1, -1] + [1] * (ndim - 1)) @ aff
        A = flipped_aff[:ndim, :ndim]
        b = flipped_aff[:ndim, -1]
        self.image.SetOrigin(b.tolist())
        if spacing2one:
            avouch(spacing is None)
            self.image.SetSpacing((1,) * ndim)
        elif spacing is not None:
            if isinstance(spacing, (int, float)): spacing = (spacing,) * ndim
            avouch(len(spacing) == ndim)
            self.image.SetSpacing(spacing)
            A = A @ np.diag(1 / np.array(spacing))
        else: A = A @ np.diag(1 / np.array(self.image.GetSpacing()))
        self.image.SetDirection(A.flatten().tolist())
    
    @property
    @alias("origin_of")
    def origin(self, image=None):
        if self.has_series:
            if isinstance(image, str): image = self.images[image]
            raise Error('Property')("No property 'origin' for micomputing.IMG with series. ")
        if image is None: image = self.image
        avouch(isinstance(image, sitk.Image), "Only Image type has origin.")
        return tuple((np.array(image.GetOrigin()) * np.array([-1, -1] + [1] * (image.GetDimension() - 2))).tolist())

    @origin.setter
    @alias("set_origin")
    def origin(self, *org):
        org = arg_tuple(org)
        if len(org) == 1: org = org * self.ndim
        if self.has_series: raise Error('Property')("Cannot set property 'origin' for micomputing.IMG with series. ")
        self.image.SetOrigin(tuple(s * x for s, x in zip([-1, -1, 1], org)))
    
    @property
    @alias("spacing_of")
    def spacing(self, image=None):
        if self.has_series:
            if isinstance(image, str): image = self.images[image]
            raise Error('Property')("No property 'spacing' for micomputing.IMG with series. ")
        if image is None: image = self.image
        avouch(isinstance(image, sitk.Image), "Only Image type has spacing.")
        return tuple(np.array(self.image.GetSpacing()).tolist())

    @spacing.setter
    @alias("set_spacing")
    def spacing(self, *sp):
        sp = arg_tuple(sp)
        if len(sp) == 1: sp = sp * self.ndim
        if self.has_series: raise Error('Property')("Cannot set property 'spacing' for micomputing.IMG with series. ")
        self.image.SetSpacing(sp)
    
    @property
    @alias("ndim_of", 'n_dim')
    def ndim(self, image=None):
        if self.has_series:
            if isinstance(image, str): image = self.images[image]
            raise Error('Property')("No property 'ndim' for micomputing.IMG with series. ")
        if image is None: image = self.image
        avouch(isinstance(image, sitk.Image), "Only Image type has ndim.")
        return self.image.GetDimension()
    
    @property
    def shape(self):
        if self.has_series: raise Error('Property')("No property 'shape' for micomputing.IMG with series. ")
        return tuple(self.image.GetSize())
    
    @property
    def orientation(self):
        if self.has_series: raise Error('Property')("No property 'orientation' for micomputing.IMG with series. ")
        return affine2orient(self.affine)
    
    def dim_of_orient_axis(self, axis):
        return dim_of_orient_axis(self.orientation, axis)

    def reorient(self, orient='LPI'):
        if self.has_series: raise Error('Property')("Cannot reorient for micomputing.IMG with series. ")
        axis = {'L':'LR', 'R':'LR', 'A':'AP', 'P':'AP', 'I':'SI', 'S':'SI'}
        from .funcs import reorient
        from .trans import Translation, Reflection, DimPermutation, ComposedTransformation
        orient_axis = [axis[i] for i in self.orientation]
        permutation = [orient_axis.index(axis[i]) for i in orient]
        if len(permutation) < len(orient_axis):
            permutation.extend(list(set(range(len(orient_axis))) - set(permutation)))
        new_orient = [self.orientation[i] for i in permutation]
        reflect_dims = [i for i, (a, b) in enumerate(zip(new_orient, orient)) if a != b]
        data = self.to_tensor()
        output_data = reorient(data, from_orient=self.orientation, to_orient=orient)
        if output_data.n_space_dim >= 3:
            output_data = bt.permute_space(output_data, 2, 1, 0, *range(3, output_data.n_space_dim))
            if output_data.n_space_dim > 3: output_data.with_sz_sequence_dim(3 - output_data.n_space_dim)
        else: output_data = output_data.transpose(0, 1)
        output_image = sitk.GetImageFromArray(output_data)
        for t in basic_tags: output_image.SetMetaData(t, self.basics[t])
        n_dim = len(permutation)
        spacing = tuple(self.image.GetSpacing()[i] for i in permutation)
        # correct_trans = Translation([(s - 1) * n for s, n in zip(spacing, data.shape)]).backward_(False)
        reflect = Reflection(reflect_dims, image_size=data).backward_(False)
        permute = DimPermutation(permutation).backward_(False)
        trans = ComposedTransformation(permute, reflect)
        affine = (np.diag([-1, -1] + [1] * (n_dim - 1)) @ self.affine @ trans.affine(n_dim).detach().cpu().numpy()).squeeze()
        output_image.SetDirection((affine[:n_dim, :n_dim] @ np.diag([1/s for s in spacing])).flatten().tolist())
        output_image.SetOrigin(affine[:n_dim, -1].tolist())
        output_image.SetSpacing(spacing)
        output = self.copy()
        output.image = output_image
        return output
    
    @property
    def patient_info(self):
        return dict(
            name = self.basics["0010|0010"],
            sex = self.basics["0010|0040"],
            age = self.basics["0010|1010"],
            height = self.basics["0010|1020"],
            weight = self.basics["0010|1030"],
            birthday = self.basics["0010|0030"]
        )
    
    @property
    def patient_name(self): return self.basics["0010|0010"]
    
    @property
    def patient_sex(self): return self.basics["0010|0040"]
    
    @property
    def patient_age(self): return self.basics["0010|1010"]
    
    @property
    def modality(self): return self.basics["0008|0060"]
    
    def resample(self, *new_spacing, new_size=None, mode='nearest'):
        if len(new_spacing) == 1 and isinstance(new_spacing[0], (tuple, list)): new_spacing = new_spacing[0]
        new_spacing = to_list(new_spacing)
        if len(new_spacing) == 1: new_spacing *= self.ndim
        if new_size is not None: new_size = to_list(new_size)
        def resample(image, new_size=new_size, mode=mode):
            resampler = sitk.ResampleImageFilter()
            if new_size is None: 
                old_spacing = image.GetSpacing()
                old_size = image.GetSize()
                new_size = [int(a * b / c) for a, b, c in zip(old_size, old_spacing, new_spacing)]
            if mode.lower() == 'linear': method = sitk.sitkLinear
            elif mode.lower() == 'nearest': method = sitk.sitkNearestNeighbor
            elif mode.lower() == 'bspline': method = sitk.sitkBSpline
            else: raise TypeError("Unrecognized argument 'mode'. ")
            resampler.SetReferenceImage(image)
            resampler.SetOutputSpacing(new_spacing)
            resampler.SetSize(new_size)
            resampler.SetTransform(sitk.Transform(3, sitk.sitkIdentity))
            resampler.SetInterpolator(method)
            return resampler.Execute(image), new_size
        if self.has_series:
            for sid in self.series_IDs:
                self.images[sid], self.series_infos[sid]['Shape'] = resample(self.images[sid])
        else: self.image, self.info['Shape'] = resample(self.image)
        return self
    
    def _compose(self, *images):
        # Not sure if this is correct. [TODO: Check]
        if len(images) == 1 and isinstance(images[0], (list, tuple)): images = images[0]
        if len(images) <= 5:
            composer = sitk.JoinSeriesImageFilter()
            try: return composer.Execute(*images)
            except: pass
        images_origin = [x.GetOrigin() for x in images]
        images_data = map(sitk.GetArrayFromImage, images)
        template = images[0]
        spacing = template.GetSpacing()
        affine = np.array(template.GetDirection())
        m = np.sqrt(len(affine))
        affine = affine.reshape((m, m))
        if len(set(images_origin)) == 1:
            output_data = np.stack(images_data)
            output_data = output_data.transpose(*range(1, output_data.ndim), 0)
        else:
            origins = np.array(images_origin)
            coords = np.linalg.inv(affine) @ origins.T
            vec_dim = item(argmax(np.abs(coords[:, 0] - coords[:, -1]).tolist()))
            spacing_z = np.sqrt(np.square(origins[1:] - origins[:-1]).sum(1)).mean()
            output_data = np.concatenate(list(images_data))
            output_data = output_data.transpose(*range(2, 4 - vec_dim), 0, *range(4 - vec_dim, output_data.ndim), 1)
            spacing = spacing[:output_data.ndim-2] + (spacing_z,) + spacing[output_data.ndim-1:]
        output_image = sitk.GetImageFromArray(output_data)
        for t in basic_tags: output_image.SetMetaData(t, self.basics[t])
        output_image.SetDirection(affine.flatten().tolist())
        output_image.SetOrigin(images_origin[0])
        output_image.SetSpacing(spacing)
        return output_image
    
    def gather_multiple_components(self):
        categories = {}
        for sid in self.series_IDs:
            segs = sid.split('.')
            if len(segs) <= 10: continue
            pid = '.'.join(segs[:10])
            categories.setdefault(pid, [])
            categories[pid].append(sid)
        for pid in categories:
            if len(categories[pid]) <= 1: continue
            images = []
            for sid in sorted(categories[pid], key=lambda x: tuple(map(eval, x.split('.')))):
                images.append(self.images.pop(sid))
                if pid not in self.series_IDs:
                    self.series_infos[pid] = self.series_infos[sid]
                    self.series_IDs.append(pid)
                self.series_IDs.remove(sid)
            self.images[pid] = self._compose(images)
            self.series_infos[pid]['Shape'] = self.images[pid].GetSize()
        self.series_IDs.sort(key=lambda x: tuple(map(eval, x.split('.'))))
    
    @property
    @alias("bundle", "get_header")
    def header(self):
        if self.has_series: ref_file = self.file_lists[0][0]
        ref_file = self.file_list[0]
        
        with run("NII"), run.use_tag:
            if self.ftype.upper() == 'NII': return nib.load(ref_file).header
        
        with run("DCM"), run.use_tag:
            if self.ftype.upper() == 'DCM': return dcm.dcmread(ref_file)

    def save_as_dtype(self, dtype: bt.dtype, path: (str, callable), as_type=None):
        if isinstance(dtype, str): dtype = getattr(bt, dtype)
        if self.has_series:
            img = self.copy()
            for sid in img.series_IDs:
                image_data = np.array(sitk.GetArrayFromImage(img.images[sid]))
                array = bt.tensor(toI(image_data)).type(dtype)
                old_image = img.images[sid]
                new_image = sitk.GetImageFromArray(array)
                new_image.SetDirection(old_image.GetDirection())
                new_image.SetSpacing(old_image.GetSpacing())
                new_image.SetOrigin(old_image.GetOrigin())
                img.images[sid] = new_image
            img.save(path, as_type=as_type)
        else:
            img = self.copy()
            image_data = np.array(sitk.GetArrayFromImage(img.image))
            array = bt.tensor(toI(image_data)).type(dtype)
            old_image = img.image
            new_image = sitk.GetImageFromArray(array)
            new_image.SetDirection(old_image.GetDirection())
            new_image.SetSpacing(old_image.GetSpacing())
            new_image.SetOrigin(old_image.GetOrigin())
            img.image = new_image
            img.save(path, as_type=as_type)

    @overload
    def save(self, data: bt.Tensor, path: (str, callable), as_type=None, **header):
        img = self.copy()
        img.from_array(data)
        for k, v in header.items():
            if not hasattr(img, k): continue
            setattr(img, k, v)
        img.save(path, as_type=as_type)

    @overload
    def save(self, img: 'IMG', path: (str, callable), as_type=None, **header):
        old_values = {}
        for k, v in header.items():
            if not hasattr(img, k): continue
            old_values[k] = getattr(img, k)
            setattr(img, k, v)
        img.save(path, as_type=as_type)
        for k, v in old_values.items(): setattr(img, k, v)

    @overload
    def save(self, trans: 'Transformation', path: (str, callable), as_type=None):
        img = self.copy()
        img.from_array(trans.toDDF(self.shape).squeeze())
        for k, v in header.items():
            if not hasattr(img, k): continue
            setattr(img, k, v)
        img.save(path, as_type=as_type)

    @overload
    def save(self, trans_file: 'TRS', path: (str, callable), as_type=None):
        img = self.copy()
        img.from_array(trans_file.trans.toDDF(self.shape).squeeze())
        for k, v in header.items():
            if not hasattr(img, k): continue
            setattr(img, k, v)
        img.save(path, as_type=as_type)

    @overload
    def save(self, path: (str, callable), as_type=None, avoid_conflict=False):
        if as_type is None:
            ftype_out = self.ftype
            if isinstance(path, str):
                path = Path(path)
                with run(sitk_supported_modalities), run.any_tag:
                    ftype_out_list = [t for e, t in sitk_modality_exts.items() if path | e]
                    if len(ftype_out_list) == 1: ftype_out = ftype_out_list[0]
                with run("DCM"), run.use_tag:
                    if path | 'dcm' or path | 'ima': ftype_out = 'DCM'
        else: ftype_out = as_type
        if ftype_out is None: raise Error("Unidentified")("Cannot save micomputing.IMG without identifing the `ftype`. ")

        if self.has_series: ids = self.series_IDs; images = self.images; infos = self.series_infos
        else: ids = [get_uid10()]; images = {ids[0]: self.image}; infos = {ids[0]: self.info}

        def get_path(sid):
            if callable(path): p = path(sid, infos)
            elif isinstance(path, str): p = Path(path)
            else: raise TypeError(f"Invalid saving path for micomputing.IMG: {path}.")
            return p
        isdir = lambda p: os.extsep not in os.path.basename(p) or len(os.path.basename(p).split(os.extsep)[-1]) > 6

        with run(sitk_supported_modalities), run.any_tag:
            if ftype_out.upper() in sitk_supported_modalities:
                ext = sitk_default_exts[ftype_out.upper()]
                used_path = []
                for sid in ids:
                    info = infos[sid]
                    p = get_path(sid)
                    if p is None: continue
                    if isdir(p):
                        if not os.path.exists(p): os.mkdir(p)
                        extra = info.copy()
                        for k in IMG.basic_info: extra.pop(k)
                        if len(extra) > 0: extra_str = ''.join([f"_{get_alphas(k)}{get_digits(str(v)) if get_digits(str(v)) else get_alphas(str(v))}" for k, v in extra.items()])
                        else: extra_str = ''
                        default_name = f"{info['SeriesTime']}_{info['SeriesDescription']}_{info['SeriesNumber']}"
                        p = os.path.join(p, default_name)
                        if p in used_path: p += extra_str
                        p //= ext
                    if p in used_path:
                        if avoid_conflict:
                            if p | ext: p = p[:-len(ext)]
                            i = 1
                            while True:
                                if p + f' ({i})' + ext in used_path: i += 1
                                else: break
                            p = p + f' ({i})' + ext
                        else: print(f"Warning: overwriting file {p} as a same path was given for multiple series. ")
                    sitk.WriteImage(images[sid], p)
                    used_path.append(p)
        with run("DCM"), run.use_tag:
            if ftype_out.upper() == 'DCM':
                used_path = []
                count = {}
                for sid in ids:
                    info = infos[sid]
                    p = get_path(sid)
                    if p is None: continue
                    if isdir(p):
                        if not os.path.exists(p): os.makedirs(p)
                        if p not in used_path: count[p] = 1
                        image = images[sid]
                        slices = []
                        ndim = image.GetDimension()
                        if ndim == 3:
                            for z in range(image.GetDepth()):
                                slices.append(image[:, :, z])
                        elif ndim == 2:
                            slices.append(image)
                        series_writer = sitk.ImageFileWriter()
                        series_writer.KeepOriginalImageUIDOn()
                        aff = self.affine_of(image)
                        orientation = np.array(image.GetDirection()).reshape((ndim, ndim)).T.flatten()[:ndim * (ndim - 1)].tolist()
                        origin = list(image.GetOrigin())[:2]
                        spacing = list(image.GetSpacing())
                        for z, image_slice in enumerate(slices):
                            slice_loc = (aff @ np.array([[0], [0], [z], [1]]))[2, 0]
                            for t, v in basic_tags.items():
                                try: image_slice.SetMetaData(t, self.basics.get(t, v))
                                except ValueError as e:
                                    raise ValueError(f"Attempting to set {t}={v}, {e}")
                            image_slice.SetMetaData("0020|0013", str(z+1))
                            image_slice.SetMetaData("0020|1041", str(-slice_loc))
                            image_slice.SetMetaData("0020|000e", sid)
                            image_slice.SetMetaData("0008|0021", cdate())
                            image_slice.SetMetaData("0008|0031", cmilitime())
                            image_slice.SetMetaData("0008|0012", cdate())
                            image_slice.SetMetaData("0008|0013", cmilitime())
                            image_slice.SetMetaData("0020|0011", str(info.get('SeriesNumber', '1')))
                            image_slice.SetMetaData("0008|103e", info.get('SeriesDescription', ''))
                            file_path = os.path.join(p, "slice%04d.dcm"%count[p])
                            series_writer.SetFileName(file_path)
                            series_writer.Execute(image_slice)
                            with dcm.dcmread(file_path) as image_again:
                                image_again.ImagePositionPatient = origin + [slice_loc] # 0020|0032
                                image_again.ImageOrientationPatient = orientation # 0020|0037
                                image_again.PixelSpacing = spacing[:2] # 0020|0030
                                image_again.SpacingBetweenSlices = str(spacing[-1]) # 0018|0088
                                image_again.SliceThickness = str(float(int(spacing[-1] - 0.1))) # 0018|0050
                                image_again.ImageType = ['DERIVED'] # 0008|0008
                                image_again.PatientSize = '200' # 0010|1020
                                image_again.save_as(file_path)
                            count[p] += 1
                        used_path.append(p)
                    else: raise TypeError(f"Please identify a directory when saving a DICOM format micomputing.IMG to '{path}'. ")
        with run("PNG"), run.use_tag:
            if ftype_out.upper() == 'PNG':
                used_path = []
                count = {}
                for sid in ids:
                    info = infos[sid]
                    p = get_path(sid)
                    if p is None: continue
                    if isdir(p):
                        if not os.path.exists(p): os.mkdir(p)
                        if p not in used_path: count[p] = 0
                        image = images[sid]
                        image = sitk.Cast(sitk.RescaleIntensity(image), sitk.sitkUInt8)
                        file_names = [os.path.join(p, "slice%04d.png"%(count[p] + z)) for z in range(image.GetDepth())]
                        count[p] += image.GetDepth()
                        series_writer = sitk.ImageSeriesWriter()
                        series_writer.SetFileNames(file_names)
                        series_writer.Execute(image)
                        used_path.append(p)
                    else: raise TypeError("Please identify a directory when saving a DICOM format micomputing.IMG. ")
                    




__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to manage package info.",
    help = "Use `info_manager(**kwargs)` to create an info object and use `with info` to check for imports. "
)

__all__ = """
    info_manager
    Hyper
    hypers
    Version
    Args
    args
""".split()

import os, re, sys, math, time, builtins, types
from optparse import OptionParser
from argparse import ArgumentParser

from .environment import get_environ_vars
from .strop import tokenize

update_format = "%Y-%m-%d %H:%M:%S"

class Version(tuple):
    def __new__(cls, string):
        matches = re.findall(r"[0-9\.]+", string)
        if len(matches) == 0: raise TypeError(f"{string} is not a version")
        return super().__new__(cls, [eval(x) for x in matches[0].split('.') if x])
    for op in "gt lt ge le eq ne".split():
        exec(f"def __{op}__(self, x): return super(Version, self).__{op}__(Version(x) if isinstance(x, str) else x)")

class info_manager:
    """
    info_manager() -> info_object

    An object storing the info of the current file.

    Note:
        It is currently provided for private use in project PyCAMIA only. 
        But it can also be used outside.

    Examples::
        Code.py:
            __info__ = info_manager(
                project="PyCAMIA", 
                package="", 
                requires="xxx"
            ).check()
            print("Project is", __info__.project)
            with __info__:
                import xxx
                from xxx import yyy
        Output:
            Project is PyCAMIA
        Error if xxx doesnot exist: ModuleNotFoundError
        Error if yyy doesnot exist in xxx: ImportError
    """
    
    @staticmethod
    def parse(code):
        info = eval(code)
        raw_args = tokenize(code, sep=['(',')'])[1]
        info.tab = raw_args[:re.search(r'\w', raw_args).span()[0]].strip('\n')
        info.order = tokenize(raw_args, sep=[',', '='], strip=None)[::2]
        return info

    def __init__(self, project = "", package = "", requires = "", **properties):
        if isinstance(requires, str): requires = requires.split()
        properties.update(dict(project=project, package=package, requires=requires))
        self.properties = properties
        self.__dict__.update(properties)
        env = get_environ_vars()
        file_path = env['__file__']
        file = os.path.extsep.join(os.path.basename(file_path).split(os.path.extsep)[:-1])
        self.name = '.'.join([x for x in [project, package, file] if x and x != "__init__"])
        self.tab = ' ' * 4
        major_keys = "project package requires".split()
        self.order = major_keys + list(set(properties.keys()) - set(major_keys))
        env['__args__'] = Args(sys.argv)
        
    def check_requires(self):
        not_found_packages = []
        for r in self.requires:
            tokens = tokenize(r, ['<', '>', '='], strip=None, keep_empty=False)
            if len(tokens) == 2: rname, rversion = tokens
            else: rname, rversion = tokens[0], None
            try:
                package = __import__(rname)
            except ModuleNotFoundError: not_found_packages.append(rname); continue
            if rversion is not None:
                op = r.replace(rname, '').replace(rversion, '').strip()
                if not eval("Version(package.__version__)" + op + "'" + rversion + "'"):
                    not_found_packages.append(r)
        if len(not_found_packages) > 0:
            raise ModuleNotFoundError(f"'{self.name}' cannot be used without dependencies {repr(not_found_packages)}.")
            
    def check(self):
        self.check_requires()
        return self
    
    def auto_install_packages(self):
        queue = self.requires.copy()
        def install(r, update=False):
            os.system(f"pip install{' --update' if update else ' '}{r}")
        while True:
            try: r = queue.pop(0)
            except IndexError: break
            tokens = tokenize(r, ['<', '>', '='], strip=None, keep_empty=False)
            if len(tokens) == 2: rname, rversion = tokens
            else: rname, rversion = tokens[0], None
            
            try: package = __import__(rname)
            except ModuleNotFoundError as error:
                error_msg = error.__str__()
                if "without dependencies" in error_msg:
                    module_names = eval(error_msg.split('[')[-1].split(']')[0])
                else: module_names = [error_msg.split("'")[1]]
                for module_name in module_names:
                    install(module_name)
                queue.append(r); continue
            if rversion is not None:
                op = r.replace(rname, '').replace(rversion, '').strip()
                if not eval("Version(package.__version__)" + op + "'" + rversion + "'"):
                    install(r, update=True); queue.append(r)
        return self
    
    def version_update(self):
        if hasattr(self, 'version'):
            version = re.sub(r"((\d+\.){2})(\d+)", lambda x: x.group(1)+str(eval(x.group(3))+1), self.version)
        else: version = '1.0.0'
        self.version = version
        self.update = time.strftime(update_format, time.localtime(time.time()))
    
    def __enter__(self): return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type == ModuleNotFoundError:
            module_name = str(exc_value).split("'")[-2]
            raise ModuleNotFoundError(f"'{self.name}' cannot be used without dependency '{module_name}'. ")
        elif exc_type == ImportError:
            module_name, _, func_name = str(exc_value).split("'")[1:4]
            raise ImportError(f"'{self.name}' requires '{func_name}' in module '{module_name}'. Please double check the version of packages. ")
        
    def __str__(self):
        args = ',\n'.join([self.tab + k + ' = ' + repr(getattr(self, k)) for k in self.order])
        return "info_manager(\n" + args + "\n)"
    
    def __getitem__(self, name):
        return getattr(self, name)
    
    def __setattr__(self, name, value):
        if hasattr(self, 'order') and name not in self.order: self.order.append(name)
        super().__setattr__(name, value)
    
    def get(self, name, value):
        if hasattr(self, name): return self[name]
        return value

class Hyper:
    def __init__(self):
        self.optParser = OptionParser()

    def add_hyper(self, *names, default=None, type=None, help_str='', **kwargs):
        if 'name' in kwargs: names += (kwargs['name'],)
        if len(names) == 0: raise TypeError("add_hyper requires a name. ")
        if type is None: type = builtins.type(default)
        if help_str == '': help_str = f"variable \'{kwargs.get('name', names[0])}\'"

        new_names = []
        for name in names:
            name = name.lstrip('-')
            abbr_match = re.match(r'\[(.)\]', name)
            if abbr_match is not None:
                abbr = abbr_match.group(1)
                name = name[:abbr_match.span()[0]] + abbr + name[abbr_match.span()[1]+1:]
                new_names.extend([f'-{abbr}', f'--{name}'])
            else:
                new_names.append(('-' if len(name) > 1 else '') + f'-{name}')

        if type == bool:
            self.optParser.add_option(
                *new_names, action = 'store_true' if default else 'store_false', 
                dest = name, default = default, help = help_str
            )
        else:
            self.optParser.add_option(
                *new_names, action = 'store', type = type, 
                dest = name, default = default, help = help_str
            )
            
    def parse_args(self, *args):
        res = self.optParser.parse_args(*args)
        self.__dict__.update(vars(res[0]))
        return res
    
    def parse_name(self, *args):
        self.parse_args(*args)
        return str(self)

    def __str__(self):
        defaults = vars(self.optParser.get_default_values())
        values = vars(self)
        values.pop('optParser')
        updated = {k: v for k, v in values.items() if k in defaults and v != defaults[k]}
        if len(updated) == 0: return "baseline"
        token = lambda k, v: str(k) + '_' + re.sub(r'\.0+$', '', str(v)[-10:])
        return '-'.join([token(k, v) for k, v in updated])

def hypers(**kwargs):
    hyper = Hyper()
    for k, v in kwargs.items():
        hyper.add_hyper(k, default=v, help_str=kwargs.get(k+'_help', ''))
    variables, args = hyper.parse_args(sys.argv)
    vs = get_environ_vars()
    vs.update(vars(variables))
    return hyper

class const:

    value = None
    
    @classmethod
    def __class_getitem__(cls, value):
        cls.value = value
        return cls

class Args:
    
    """
    Argument manager for a python script. An enclosure of argparse.ArgumentParser. 

    Example::
        >>> args = Args("a.py -fv --n_epoch 4 --n_batch 3".split()[1:])
        >>> args.n_epoch = 5
        >>> args.n_batch[int: "The batch size for the network training. "] = 5
        >>> args.gpu[list[2]: "The 2 GPU ids. "] = [2, 3]
        >>> with args('-f'): args.flag[bool: "A flag argument which is False if --flag is not given. "] = False
        >>> with args('-v', '--use_valid'): args.n_iter_each_valid[Args.const[10]: "Whether to use validation or not. "] = int(1e6)
        >>> args
        Args(n_epoch=4, n_batch=3, gpu=[2, 3], flag=True, n_iter_each_valid=10)
        >>> args.n_batch
        3
        >>> print(args.help())
        usage: <code> [-h] [--n_epoch N_EPOCH] [--n_batch N_BATCH] [--gpu GPU GPU] [-f [FLAG]] [-v [N_ITER_EACH_VALID]]

        options:
        -h, --help            show this help message and exit
        --n_epoch N_EPOCH     The maximal number of epochs for training.
        --n_batch N_BATCH     The batch size for the network training.
        --gpu GPU GPU         The 2 GPU ids.
        -f [FLAG]             A flag argument which is False if --flag is not given.
        -v [N_ITER_EACH_VALID], --use_valid [N_ITER_EACH_VALID]
                                Whether to use validation or not.
    """
    
    default_help_strs = dict(
        n_batch = "Batch size for each iteration. ", 
        batch_size = "Batch size for each iteration. ", 
        n_epoch = "The maximal number of epochs for training. ", 
        epochs = "The maximal number of epochs for training. ", 
        max_epoch = "The maximal number of epochs for training. ", 
        init_epoch = "The index of epoch for the first iteration, used in restarts of training. ",
        i_gpu = "The id of GPU(s)", 
        gpu = "The id of GPU(s)", 
    )
    
    const = const
    
    @staticmethod
    def tags(self, name):
        abbr_match = re.match(r'\[(.)\]', name)
        if abbr_match is not None:
            abbr = abbr_match.group(1)
            name = name[:abbr_match.span()[0]] + abbr + name[abbr_match.span()[1]+1:]
            names=[f'-{abbr}', f'--{name.lstrip("-")}']
        else:
            names=[('-' if len(name) > 1 else '') + f'-{name}']
        return names
    
    class Arg:
        def __init__(self, parent, name):
            self.parent = parent
            self.name = name
            
        def __setitem__(self, subscript, value):
            if isinstance(subscript, str):
                _type = value.__class__
                _help = subscript
            elif isinstance(subscript, type):
                _type = subscript
                _help = Args.default_help_strs.get(self.name, f"Argument '{self.name}' for {self.parent.prog}. ")
            elif isinstance(subscript, slice):
                if subscript.step is not None: raise TypeError("Only one colon is allowed in argument expression. ")
                _type = subscript.start
                _help = subscript.stop
            if _type == Args.const: nargs = '?'; const = _type.value; _type = type(_type.value)
            elif _type == bool:
                nargs = '?'
                if isinstance(value, bool): const = not value
                else: const = True
            else:
                if not isinstance(_type, (type, types.GenericAlias)): raise TypeError("Only types or Args.const is allowed before colon. ")
                if issubclass(getattr(_type, '__origin__', _type), (tuple, list)) and isinstance(_type, types.GenericAlias):
                    nargs = _type.__args__[0]
                    if isinstance(nargs, type):
                        _type = _type.__origin__[nargs]
                        nargs = None
                    else:
                        if isinstance(nargs, slice): nargs = '+'
                        nargs = {'>0': '+', ...: '*', -1: '*'}.get(nargs, nargs)
                        _type = _type.__origin__
                elif str(_type.__class__) == "<class 'pyoverload.typings.ArrayType'>":
                    nargs = len(_type)
                    if nargs < 0: nargs = '*'
                    _type = _type.origin_type[_type.base_type[0]]
                else:
                    nargs = (len(value) if len(value) > 0 else '*') if isinstance(value, (tuple, list)) else None
                const = None
            tags = self.parent.additional_tags
            if not tags: tags = ['--' + self.name]
            dest = self.name
            # if not tags: tags = ['--' + self.name]
            # if '--' + self.name not in tags: dest = self.name
            # else: dest = None
            self.parent.add_argument(*tags, nargs=nargs, const=const, dest=dest, type=_type, default=value, help=_help)
            if not self.parent.parse_attempt(key=self.name):
                setattr(self.parent, self.name, value)
            
    class Scope:
        
        def __init__(self, parent, *tags):
            if len(tags) == 1 and isinstance(tags[0], (tuple, list)): tags = tags[0]
            self.tags = list(tags)
            self.parent = parent
            
        def __enter__(self): self.parent.additional_tags = self.tags
        def __exit__(self, type, value, trace): self.parent.additional_tags = []
        
    class Argument:
        
        def __init__(self, *tags, nargs=1, const=None, dest='', type=None, default=None, help=''):
            self.name = dest
            fullnames = []
            abbrs = []
            for tag in tags:
                if tag.startswith('--'): fullnames.append(tag.lstrip('-'))
                elif tag.startswith('-'): abbrs.append(tag.lstrip('-'))
                else: raise TypeError(f"Invalid argument tag '{tag}' in '{tags}' for variable {dest}. ")
            if len(set(abbrs)) < len(abbrs): raise TypeError(f"Duplicated abbreviation in {abbrs}. ")
            if len(fullnames) > 0:
                found_abr = {}
                for abr in abbrs:
                    for name in fullnames:
                        if name in found_abr: continue
                        if abr.lower() in name.lower(): found_abr[name] = [abr]; break
                    else:
                        for name in fullnames:
                            if name not in found_abr: continue
                            if abr.lower() in name.lower(): found_abr[name].append(abr); break
                        else: raise TypeError(f"Inappropriate abbreviation '{abr}' as it is not in '{'/'.join(fullnames)}'. ")
                def mark_name(name, chars):
                    for ch in chars:
                        i = name.lower().index(ch.lower())
                        name = name[:i] + '[' + ch + ']' + name[i+1:]
                    return name
                self.tag_label = '--' + '/'.join(name if name not in found_abr else mark_name(name, found_abr[name]) for name in fullnames)
            else: self.tag_label = '-' + ''.join(abbrs)
            self.tags = tags
            self.nargs = 1 if nargs is None else nargs
            self.const = const
            self.dest = dest
            self.type = builtins.type(default) if type is None else type
            self.default = default
            self.help = help
        
        @property
        def optional(self): return self.default is not None
        
        def __repr__(self): return f"<Argument {self.name}[{self.type.__name__}]: {'/'.join(self.tags)}>"

    properties = ['argv', 'prog', 'description', 'arguments', 'arg_names', 'additional_tags', 'end_parse', 'calling_help']
    
    def __init__(self, argv=None, prog='', description=''):
        if not prog:
            if argv is None: prog, *argv = sys.argv
            else: prog = '<code>'
        elif argv is None: _, *argv = sys.argv
        self.argv = argv
        self.prog = prog
        self.description = description
        
        self.arguments = [Args.Argument('-h', '--help', nargs='?', dest='help', help="Show this help message and exit. ")]
        self.arg_names = []
        self.additional_tags = []
        self.end_parse = False
        self.calling_help = False
    
    def __getattribute__(self, name):
        try: return super().__getattribute__(name)
        except AttributeError:
            if self.end_parse: raise AttributeError(f"Argument '{name}' is not defined. ")
            return Args.Arg(self, name)
    
    def __setattr__(self, name, value):
        if name in Args.properties: return super(Args, self).__setattr__(name, value)
        if name in self.arg_names: return super(Args, self).__setattr__(name, value)
        help_str = Args.default_help_strs.get(name, f"Argument '{name}' for {self.prog}. ")
        tags = self.additional_tags
        if not tags: tags = ['--' + name]
        dest = name
        # if not tags: tags = ['--' + name]
        # if '--' + name not in tags: dest = name
        # else: dest = None
        self.add_argument(*tags, dest=dest, type=type(value), default=value, help=help_str)
        if not self.parse_attempt(key=name):
            return super(Args, self).__setattr__(name, value)
        
    def __call__(self, *tags):
        return Args.Scope(self, *tags)
    
    def __enter__(self): return self
    
    def __exit__(self, error_type, error_str, traceback):
        if error_type is not None:
            raise error_type(error_str)
        self.fix_arguments()
        
    @property
    def usage(self):
        flags = []
        fixeds = []
        optionals = []
        for arg in self.arguments:
            if not arg.tag_label.startswith('--') and arg.nargs in ('?', 0):
                flags.append(arg.tag_label.lstrip('-')); continue
            optional = arg.optional
            if arg.nargs == '?': optional = True; nargs = 0
            elif arg.nargs in ('+', '*'): nargs = 1
            else: nargs = arg.nargs
            arg_usage = ' '.join([arg.tag_label] + [arg.name.upper()] * nargs)
            if arg.nargs in ('+', '*'): arg_usage += arg.nargs
            if optional: optionals.append(f'[{arg_usage}]')
            else: fixeds.append(arg_usage)
        fregs = [self.prog]
        fregs.extend(fixeds)
        if flags: fregs.append('-' + ''.join(flags))
        fregs.extend(optionals)
        return ' '.join(fregs)
    
    @property
    def help_str(self, indent=2, usage_width=20, help_width=90):
        el = '\n'; usage_indent = ' ' * indent; help_indent = ' ' * (indent * 2 + usage_width)
        string = "usage: " + self.usage + el*2
        string += "options: " + el
        for arg in self.arguments:
            optional = "; optional" if arg.optional else ''
            if arg.nargs == '?': optional = "; flag"; nargs = 0
            elif arg.nargs in ('+', '*'): nargs = 1
            else: nargs = arg.nargs
            arg_usages = ', '.join(' '.join([tag] + [arg.name.upper()] * nargs) + (arg.nargs if arg.nargs in ('+', '*') else '') for tag in arg.tags)
            if arg.name != 'help': help_content = f"[{arg.type.__name__}{optional}] " + arg.help.strip() + f" Defaults to {arg.default!r}. "
            else: help_content = arg.help
            help_lines = []
            ipointer = 0
            while True:
                try: last_space = help_content.rindex(' ', ipointer, ipointer + help_width)
                except ValueError:
                    line = help_content[ipointer:]
                    if line.strip(): help_lines.append(line)
                    break
                line = help_content[ipointer:last_space]
                if line.strip(): help_lines.append(line)
                ipointer = last_space + 1
            help_string = (el+help_indent).join(help_lines)
            help_string += el
            if len(arg_usages) > usage_width:
                string += usage_indent + arg_usages + el
                string += help_indent + help_string
            else: string += usage_indent + arg_usages + ' ' * (usage_width - len(arg_usages) + indent) + help_string
        return string
    
    def register(self, *tags, nargs=0, const=None, dest='', type=None, default=None, help=''):
        if not dest: raise TypeError(f"Undecided variable name for '{tags}'")
        self.arguments.append(Args.Argument(*tags, nargs=nargs, const=const, dest=dest, type=type, default=default, help=help))
        self.arg_names.append(dest)
    
    def add_argument(self, *tags, nargs=0, const=None, dest='', type=None, default=None, help=''):
        if not dest: raise TypeError(f"Undecided variable name for '{tags}'")
        self.arguments.append(Args.Argument(*tags, nargs=nargs, const=const, dest=dest, type=type, default=default, help=help))
        self.arg_names.append(dest)
    
    def parse_attempt(self, input_argv=None, key=None):
        # self.calling_help = True
        # return False
        if self.calling_help: return False
        if input_argv is None: input_argv = self.argv
        flags = []
        for arg in self.arguments:
            if not arg.tag_label.startswith('--') and arg.nargs in ('?', 0):
                flags.append(arg.tag_label.lstrip('-'))
        key_found = False
        section = []
        cur_tag = None
        for va in input_argv + ['-']:
            if not va.startswith('-'):
                if cur_tag is not None: section.append(va)
                continue
            if cur_tag is not None:
                # Find the argument. 
                if not cur_tag.startswith('--'): tags = ['-' + t for t in cur_tag[1:]]
                else: tags = [cur_tag]
                for tag in tags:
                    for arg in self.arguments:
                        if tag in arg.tags: break
                    else:
                        # Failed to find argument. 
                        if key is None: return False
                        continue
                    # Argument found, proceed. 
                    if key is not None and arg.name == key: key_found = True
                    if arg.name == 'help': self.calling_help = True; return False
                    if isinstance(arg.nargs, int) and arg.nargs > 1 or arg.nargs in ('+', '*'):
                        try:
                            if issubclass(getattr(arg.type, '__origin__', arg.type), (tuple, list)) and isinstance(arg.type, types.GenericAlias):
                                subtype = arg.type.__args__[0]
                                value = arg.type.__origin__([subtype(x) for x in section])
                            else: value = arg.type(section)
                        except ValueError: raise TypeError(f"Error in casting {' '.join(section)} to type {arg.type} for argument {arg.name}: {tag} {' '.join(section)}. ")
                        if isinstance(arg.nargs, int) and len(value) != arg.nargs:
                            raise TypeError(f"Error in assigning {value} (of length {len(value)}) to argument {arg.name}: {tag} {' '.join(section)}. ")
                    else:
                        value_str = ' '.join(section)
                        try: value = arg.type(value_str)
                        except ValueError: raise TypeError(f"Error in casting {value_str} to type {arg.type} for argument {arg.name}: {tag} {value_str}. ")
                    setattr(self, arg.name, value)
            cur_tag = va
            section = []
        if key is not None: return key_found
        return True
    
    def complete(self):
        self.fix_arguments()
    
    def fix_arguments(self):
        self.end_parse = True
        if self.calling_help:
            print(self.help_str)
            exit(0)
    
    def help(self):
        return self.help_str
    
    def __str__(self):
        return f"Args({', '.join(f'{a}={getattr(self, a)!r}' for a in self.arg_names)})"
    
    __repr__ = __str__
    
args = Args()




try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "Efficient typehint of input argument type for functions. ",
    requires = "pycamia"
)

__all__ = """
    typehint
    TypeHintError
    HintTypeError
    
    get_func_info
    get_arg_values
    get_virtual_declaration
    
    deprecated params
""".split()

deprecated = ...
def params(*_a, **_k): raise NotImplementedError("""
[WARNING] pyoverload has upgraded into pyoverload II, please read the new README.md for detailed information.
To simply get your previous codes running, do this: 
from pyoverload.old_version_files_deprecated.override import overload
from pyoverload.old_version_files_deprecated.typehint import *
If you are using the 'decorator fashion' (@overload for all imp's) of pyoverload I, use the following imports of new functions will work.
from pyoverload import overload
from pyoverload import typehint as params
from pyoverload.old_version_files_deprecated.typehint import * # Special types such as Int, etc. still need this line of import. 
Do change the capitalized types into lower_cased ones, and @params into @typehint later.
""")

from .typings import *
from functools import wraps

def decorator(wrapper_func):
    if not isinstance(wrapper_func, functional):
        raise TypeError(f"@decorator wrapping a non-wrapper: {wrapper_func.__qualname__}")
    def wrapper(*args, **kwargs):
        if (len(args) == 1 and isinstance(args[0], functional) or # functions
            len(args) == 2 and isinstance(args[1], functional)): # methods
            func = args[-1]
            raw_func = func.__func__ if isinstance(func, method) else func
            raw_func = getattr(raw_func, '__wrapped__', raw_func)
            func_name = f"{raw_func.__name__}:{wrapper_func.__name__}"
            outer_func = wraps(raw_func)(wrapper_func(*args, **kwargs))
            outer_func.__name__ = func_name
            outer_func.__doc__ = raw_func.__doc__
            if isinstance(func, staticmethod): trans = staticmethod
            elif isinstance(func, classmethod): trans = classmethod
            else: trans = lambda x: x
            return trans(outer_func)
        return decorator(wrapper_func(*args, **kwargs))
        # raise TypeError(f"Invalid decorator '@{wrapper_func.__qualname__}' for {args}")
    return wraps(wrapper_func)(wrapper)

class TypeHintError(Exception): ...
class HintTypeError(Exception): ...

def get_annot(annotation: (type, tuple, str, None, list, dict)):
    # type, tuple, str, None or list for *args, dict for **kwargs
    if isinstance(annotation, list): return '[' + ', '.join(get_annot(a) for a in annotation) + ']'
    if isinstance(annotation, dict): return '{' + ', '.join(repr(str(k)) + ': ' + get_annot(a) for k, a in annotation.items()) + '}'
    if annotation is ...: return '...'
    if annotation is None: return 'None'
    if isinstance(annotation, str): return repr(annotation)
    if isinstance(annotation, tuple):
        if len(annotation) >= 2: return '(' + ', '.join(get_annot(a) for a in annotation) + ')'
        if len(annotation) == 1: return '(' + get_annot[annotation[0]] + ',)'
        if len(annotation) == 0: return 'tuple()'
    return getattr(annotation, '__name__', str(annotation))

def get_virtual_declaration(func: functional):
    is_method = False
    if isinstance(func, method):
        is_method = True
        while hasattr(func, '__func__'): func = func.__func__
    annotations = func.__annotations__
    f_code = func.__code__
    f_name = func.__name__
    f_defaults = [] if func.__defaults__ is None else func.__defaults__
    f_kwdefaults = {} if func.__kwdefaults__ is None else func.__kwdefaults__
    co_nonvarargcount = f_code.co_argcount + f_code.co_kwonlyargcount
    parts = []
    for i, v in enumerate(f_code.co_varnames):
        if i == f_code.co_posonlyargcount and f_code.co_posonlyargcount > 0: parts.append('/')
        if i == f_code.co_argcount:
            if f_code.co_flags & 0x04:
                argv = f_code.co_varnames[co_nonvarargcount]
                if argv in annotations:
                    term = argv + ': ' + get_annot(annotations[argv])
                else: term = argv
                parts.append('*' + term)
            else: parts.append('*')
        if i >= co_nonvarargcount:
            if f_code.co_flags & 0x08:
                has_args = (f_code.co_flags & 0x04) >> 2
                kwargv = f_code.co_varnames[co_nonvarargcount + has_args]
                if kwargv in annotations:
                    term = kwargv + ': ' + get_annot(annotations[kwargv])
                else: term = kwargv
                parts.append('**' + term)
            break
        if i == 0 and is_method: continue
        if v in annotations:
            term = v + ': ' + get_annot(annotations[v])
        else: term = v
        j = i + len(f_defaults) - f_code.co_argcount
        if 0 <= j < len(f_defaults):
            term += '= ' + repr(f_defaults[j])
        if i >= f_code.co_argcount and v in f_kwdefaults:
            term += '= ' + repr(f_kwdefaults[v])
        parts.append(term)
    dec = f"{f_name}({', '.join(parts)})"
    if is_method: dec = '_.' + dec
    if 'return' in annotations:
        dec += ' -> ' + get_annot(annotations['return'])
    return dec

def get_arg_values(func_info: tuple, *args, **kwargs):
    (   f_name, has_args, has_kwargs, f_varnames,
        co_posonlyargcount, co_argcount, co_nonvarargcount, co_varcount,
        n_defaults, f_defaults, f_kwdefaults
    ) = func_info
    
    n_args = len(args)
    
    if has_kwargs:
        s = co_posonlyargcount
        f_varnames = f_varnames[s:]
        values = list(args[:co_posonlyargcount])
    else:
        s = 0
        values = []
    for i, v in zip(range(s, co_nonvarargcount), f_varnames):
        try:
            value = kwargs.pop(v, None)
            if i < co_posonlyargcount and value is not None:
                raise HintTypeError(f"{f_name}() got some positional-only arguments passed as keyword arguments: '{v}'")
            if i < n_args and i < co_argcount and value is not None:
                raise HintTypeError(f"{f_name}() got multiple values for argument '{v}'")
            if value is None:
                value = f_kwdefaults[v] if co_argcount <= i else (args[i] if i < n_args else f_defaults[i - co_argcount + n_defaults])
        except IndexError:
            missing_args = f_varnames[n_args:co_argcount - n_defaults]
            n_missing = len(missing_args)
            missing_args_str = ', '.join([repr(x) for x in missing_args[:-1]])
            missing_args_str += (' and ' if missing_args_str else '') + repr(missing_args[-1])
            raise HintTypeError(f"{f_name}() missing {n_missing} required position argument{'s' if n_missing > 1 else ''}: {missing_args_str}")
        except KeyError:
            missing_args = [v for v in f_varnames[co_argcount:co_nonvarargcount] if v not in f_kwdefaults]
            n_missing = len(missing_args)
            missing_args_str = ', '.join([repr(x) for x in missing_args[:-1]])
            missing_args_str += (' and ' if missing_args_str else '') + repr(missing_args[-1])
            raise HintTypeError(f"{f_name}() missing {n_missing} required keyword-only argument{'s' if n_missing > 1 else ''}: {missing_args_str}")
        values.append(value)
    if has_args:
        values.append(args[co_argcount:])
    elif n_args > co_argcount:
        raise HintTypeError(f"{f_name}() takes {co_argcount} positional argument but {n_args} were given")
    if has_kwargs: values.append(kwargs)
    elif len(kwargs) > 0:
        n_unexpkw = len(kwargs)
        unexpected_kw = list(kwargs.keys())
        if n_unexpkw == 1: unexpected_kw = repr(unexpected_kw[0])
        else: unexpected_kw = ' and '.join([', '.join([repr(x) for x in unexpected_kw[:-1]]), repr(unexpected_kw[-1])])
        raise HintTypeError(f"{f_name}() got{' an ' if n_unexpkw == 1 else ' '}unexpected keyword{'s' if n_unexpkw > 1 else ''} argument {unexpected_kw}")
    return values

def check_annotations(func_info: tuple, values: list, annotations: list):
    (   f_name, has_args, has_kwargs, f_varnames,
        co_posonlyargcount, co_argcount, co_nonvarargcount, co_varcount,
        n_defaults, f_defaults, f_kwdefaults
    ) = func_info

    for i, (v, a) in enumerate(zip(values, annotations)):
        if a is None: continue
        if i < co_nonvarargcount:
            if isinstance(a, (type, tuple, str)):
                if not isinstance(v, a):
                    raise TypeHintError(f"{f_name}() needs argument '{f_varnames[i]}' of type {get_annot(a)}, but got {repr(v)} of type {get_annot(v.__class__)}")
            else: raise HintTypeError(f"Invalid annotation for '{f_varnames[i]}': {get_annot(a)}")
        elif i == co_nonvarargcount and has_args:
            if isinstance(a, list):
                if ... in a:
                    if not len(a) <= len(v) + 1: raise TypeHintError(f"Variable-arguments (*{f_varnames[i]}) argument fewer thatn listed typehint annotation. Annotation of length {len(a)} but input of length {len(v)}")
                    i_ellipsis = [i for i, t in enumerate(a) if t == ...]
                    if len(i_ellipsis) > 1: raise TypeHintError(f"List typehint annotation for variable-arguments (*{f_varnames[i]}) argument can only have a single ellipsis '...' for omitted types with arbitrary length. Use None for others")
                    epsi = i_ellipsis[0]
                    if not all(isinstance(x, y) for x, y in zip(v[:epsi], a[:epsi])) or not all(isinstance(x, y) for x, y in zip(v[len(v) - len(a) + epsi + 1:], a[epsi+1:])):
                        raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[i]}' of types {get_annot(a)}, but got {get_annot([x.__class__ for x in v])}: {v}")
                elif len(a) != len(v):
                    i_ellipsis = [i for i, t in enumerate(a) if getattr(t, 'base_size', None) == (...,)]
                    if len(i_ellipsis) != 1: raise TypeHintError(f"List typehint annotation for variable-arguments (*{f_varnames[i]}) argument needs to be the same size of the input. Annotation of length {len(a)} but input of length {len(v)}, use ... for omitted types")
                    mid_t = a[i_ellipsis[0]]
                    if mid_t.origin_type == iterable:
                        mid_t = mid_t.base_type
                    else: mid_t = mid_t.origin_type
                    if (
                        not all(isinstance(x, y) for x, y in zip(v[:epsi], a[:epsi])) or 
                        not all(isinstance(x, y) for x, y in zip(v[len(v) - len(a) + epsi + 1:], a[epsi+1:])) or
                        not all(isinstance(x, mid_t) for x in v[epsi:len(v) - len(a) + epsi + 1])
                    ):
                        raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[i]}' of types {get_annot(a)}, but got {get_annot([x.__class__ for x in v])}: {v}")
                if not all(isinstance(x, y) for x, y in zip(v, a)):
                    raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[i]}' of types {get_annot(a)}, but got {get_annot([x.__class__ for x in v])}: {v}")
            else:
                if not all(isinstance(x, a) for x in v):
                    raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[i]}' with all elements of type {get_annot(a)}, but got {get_annot([x.__class__ for x in v])}: {v}")
        elif has_kwargs:
            if isinstance(a, dict):
                if not all(isinstance(v[k], a[k]) for k in v if k in a):
                    k = [k for k in v if k in a and not isinstance(v[k], a[k])][0]
                    raise TypeHintError(f"{f_name}() needs keyword argument '{k}' (in '**{f_varnames[i]}') of type {get_annot(a[k])}, but got {repr(v[k])} of type {get_annot(v[k].__class__)}")
            else:
                if not all(isinstance(x, a) for x in v.values()):
                    k = [k for k, x in v.items() if not isinstance(x, a)][0]
                    raise TypeHintError(f"{f_name}() needs keyword argument '{k}' (in '**{f_varnames[i]}') of type {get_annot(a)}, but got {repr(v[k])} of type {get_annot(v[k].__class__)}")

def match_and_check(func_info: tuple, annotations: list, *args, **kwargs):
    (   f_name, has_args, has_kwargs, f_varnames,
        co_posonlyargcount, co_argcount, co_nonvarargcount, co_varcount,
        n_defaults, f_defaults, f_kwdefaults
    ) = func_info
    
    n_args = len(args)
    
    for i, v, a in zip(range(co_nonvarargcount), f_varnames, annotations):
        
        if a is None: continue
        if not isinstance(a, (type, tuple, str)): raise HintTypeError(f"Invalid annotation for '{v}': {get_annot(a)}")
        try:
            if has_kwargs and i < co_posonlyargcount and i < n_args:
                if not isinstance(args[i], a):
                    raise TypeHintError(f"{f_name}() needs argument '{v}' of type {get_annot(a)}, but got {repr(args[i])} of type {get_annot(args[i].__class__)}")
                continue
            value = kwargs.pop(v, None)
            if i < co_posonlyargcount and value is not None:
                raise HintTypeError(f"{f_name}() got some positional-only arguments passed as keyword arguments: '{v}'")
            if i < n_args and i < co_argcount and value is not None:
                raise HintTypeError(f"{f_name}() got multiple values for argument '{v}'")
            if value is None:
                value = f_kwdefaults[v] if co_argcount <= i else (args[i] if i < n_args else f_defaults[i - co_argcount + n_defaults])
        except IndexError:
            missing_args = f_varnames[n_args:co_argcount - n_defaults]
            n_missing = len(missing_args)
            missing_args_str = ', '.join([repr(x) for x in missing_args[:-1]])
            missing_args_str += (' and ' if missing_args_str else '') + repr(missing_args[-1])
            raise HintTypeError(f"{f_name}() missing {n_missing} required position argument{'s' if n_missing > 1 else ''}: {missing_args_str}")
        except KeyError:
            missing_args = [v for v in f_varnames[co_argcount:co_nonvarargcount] if v not in f_kwdefaults]
            n_missing = len(missing_args)
            missing_args_str = ', '.join([repr(x) for x in missing_args[:-1]])
            missing_args_str += (' and ' if missing_args_str else '') + repr(missing_args[-1])
            raise HintTypeError(f"{f_name}() missing {n_missing} required keyword-only argument{'s' if n_missing > 1 else ''}: {missing_args_str}")
        except Exception as error:
            f_varnames = list(f_varnames)
            var_names = f_varnames[:co_posonlyargcount] + ['/'] + f_varnames[co_posonlyargcount:co_argcount]
            var_names += ['*' + ('' if not has_args else f_varnames[co_nonvarargcount])]
            var_names += f_varnames[co_argcount:co_nonvarargcount]
            if has_kwargs: var_names += ['**' + f_varnames[co_nonvarargcount + has_args]]
            i = f_varnames.index(v)
            var_names[i] = var_names[i] + ': ' + get_annot(a) + ' = ' + repr(value)
            args_str = ', '.join(var_names)
            raise error.__class__(f"In checking {f_name}({args_str}): " + error.__str__())
        if not isinstance(value, a):
            raise TypeHintError(f"{f_name}() needs argument '{v}' of type {get_annot(a)}, but got {repr(value)} of type {get_annot(value.__class__)}")
        continue

    if has_args:
        a = annotations[co_nonvarargcount]
        if a is not None:
            if isinstance(a, list):
                if ... in a:
                    if not len(a) <= n_args - co_argcount + 1: raise TypeHintError(f"Variable-arguments (*{f_varnames[co_nonvarargcount]}) argument fewer than annotated types. Annotation of length {len(a)} but unassigned input of length {n_args - co_argcount}")
                    i_ellipsis = [i for i, t in enumerate(a) if t == ...]
                    if len(i_ellipsis) > 1: raise TypeHintError(f"List typehint annotation for variable-arguments (*{f_varnames[co_nonvarargcount]}) argument can only have a single ellipsis '...' for omitted types with arbitrary length. Use None for others")
                    epsi = i_ellipsis[0]
                    if not all(isinstance(x, y) for x, y in zip(args[co_argcount:][:epsi], a[:epsi])) or not all(isinstance(x, y) for x, y in zip(args[co_argcount:][n_args - co_argcount - len(a) + epsi + 1:], a[epsi+1:])):
                        raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[co_nonvarargcount]}' of types {get_annot(a)}, but got {get_annot([x.__class__ for x in args[co_argcount:]])}: {args[co_argcount:]}")
                elif len(a) != n_args - co_argcount:
                    i_ellipsis = [i for i, t in enumerate(a) if getattr(t, 'base_size', None) == (...,)]
                    if len(i_ellipsis) != 1: raise TypeHintError(f"Variable-arguments (*{f_varnames[co_nonvarargcount]}) argument has a different length from the annotated types. Annotation of length {len(a)} but unassigned input of length {n_args - co_argcount}, use ... for omitted types")
                    epsi = i_ellipsis[0]
                    mid_t = a[epsi]
                    if mid_t.origin_type == iterable:
                        mid_t = mid_t.base_type
                    else: mid_t = mid_t.origin_type
                    if (
                        not all(isinstance(x, y) for x, y in zip(args[co_argcount:][:epsi], a[:epsi])) or 
                        not all(isinstance(x, y) for x, y in zip(args[co_argcount:][n_args - co_argcount - len(a) + epsi + 1:], a[epsi+1:])) or
                        not all(isinstance(x, mid_t) for x in args[co_argcount:][epsi:n_args - co_argcount - len(a) + epsi + 1])
                    ):
                        raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[co_nonvarargcount]}' of types {get_annot(a)}, but got {get_annot([x.__class__ for x in args[co_argcount:]])}: {args[co_argcount:]}")
            elif not isinstance(a, (type, tuple, str)): raise HintTypeError(f"Invalid annotation for '{f_varnames[co_nonvarargcount]}': {get_annot(a)}")
            elif not all(isinstance(x, a) for x in args[co_argcount:]):
                raise TypeHintError(f"{f_name}() needs variable-arguments '*{f_varnames[co_nonvarargcount]}' with all elements of type {get_annot(a)}, but got {get_annot([x.__class__ for x in args[co_argcount:]])}: {args[co_argcount:]}")
    elif n_args > co_argcount:
        raise HintTypeError(f"{f_name}() takes {co_argcount} positional argument but {n_args} were given")
    if has_kwargs:
        a = annotations[co_nonvarargcount + has_args]
        if a is not None:
            if isinstance(a, dict):
                if not all(isinstance(kwargs[k], a[k]) for k in kwargs if k in a):
                    v = f_varnames[co_nonvarargcount + has_args]
                    k = [k for k in kwargs if k in a and not isinstance(kwargs[k], a[k])][0]
                    raise TypeHintError(f"{f_name}() needs keyword argument '{k}' (in '**{v}') of type {get_annot(a[k])}, but got {repr(kwargs[k])} of type {get_annot(kwargs[k].__class__)}")
                return
            if not isinstance(a, (type, tuple, str)): raise HintTypeError(f"Invalid annotation for '{f_varnames[co_nonvarargcount + has_args]}': {get_annot(a)}")
            if not all(isinstance(x, a) for x in kwargs.values()):
                v = f_varnames[co_nonvarargcount + has_args]
                k = [k for k, x in kwargs.items() if not isinstance(x, a)][0]
                raise TypeHintError(f"{f_name}() needs keyword argument '{k}' (in '**{v}') of type {get_annot(a)}, but got {repr(kwargs[k])} of type {get_annot(kwargs[k].__class__)}")
    else:
        for v in f_varnames[co_posonlyargcount:]: kwargs.pop(v, None)
        if len(kwargs) > 0:
            n_unexpkw = len(kwargs)
            unexpected_kw = list(kwargs.keys())
            if n_unexpkw == 1: unexpected_kw = repr(unexpected_kw[0])
            else: unexpected_kw = ' and '.join([', '.join([repr(x) for x in unexpected_kw[:-1]]), repr(unexpected_kw[-1])])
            raise HintTypeError(f"{f_name}() got{' an ' if n_unexpkw == 1 else ' '}unexpected keyword{'s' if n_unexpkw > 1 else ''} argument {unexpected_kw}")

def v_args(func):
    return func.__code__.co_varnames[func.__code__.co_argcount + func.__code__.co_kwonlyargcount] if func.__code__.co_flags & 0x04 else None

def v_kwargs(func):
    return func.__code__.co_varnames[func.__code__.co_argcount + func.__code__.co_kwonlyargcount + ((func.__code__.co_flags & 0x04) >> 2)] if func.__code__.co_flags & 0x08 else None

def get_func_info(func):
    if isinstance(func, method): func = func.__func__
    wrapped_func = func
    while hasattr(wrapped_func, '__wrapped__'): wrapped_func = wrapped_func.__wrapped__
    f_name = wrapped_func.__name__
    f_code = wrapped_func.__code__
    has_args = (f_code.co_flags & 0x04) >> 2
    has_kwargs = (f_code.co_flags & 0x08) >> 3
    co_posonlyargcount = f_code.co_posonlyargcount
    co_argcount = f_code.co_argcount
    co_nonvarargcount = f_code.co_argcount + f_code.co_kwonlyargcount # co_nonvarargcount
    co_varcount = f_code.co_argcount + f_code.co_kwonlyargcount + has_args + has_kwargs # co_varcount
    f_varnames = f_code.co_varnames
    f_defaults = tuple() if wrapped_func.__defaults__ is None else wrapped_func.__defaults__
    n_defaults = len(f_defaults)
    f_kwdefaults = {} if wrapped_func.__kwdefaults__ is None else wrapped_func.__kwdefaults__
    return (
        f_name, has_args, has_kwargs, f_varnames,
        co_posonlyargcount, co_argcount, co_nonvarargcount, co_varcount,
        n_defaults, f_defaults, f_kwdefaults
    )

@decorator
def typehint(*types, **kwtypes):
    # Usage 1: @typehint(*, check_return=False, check_annot_only=False)
    if len(types) == 0 and (
        len(kwtypes) == 1 and ('check_return' in kwtypes or 'check_annot_only' in kwtypes) or
        len(kwtypes) == 2 and ('check_return' in kwtypes and 'check_annot_only' in kwtypes)
    ):
        return lambda f: typehint(f, **kwtypes)
    # Usage 2: func = typehint(func, /, *, annotations=None, check_return=False, check_annot_only=False) or @typehint
    elif len(types) == 1 and isinstance(types[0], functional) and (
        len(kwtypes) == 0 or
        len(kwtypes) == 1 and ('annotations' in kwtypes or 'check_return' in kwtypes or 'check_annot_only' in kwtypes) or
        len(kwtypes) == 2 and ('annotations' in kwtypes and 'check_return' in kwtypes or
                               'annotations' in kwtypes and 'check_annot_only' in kwtypes or
                               'check_return' in kwtypes and 'check_annot_only' in kwtypes) or
        len(kwtypes) == 3 and ('annotations' in kwtypes and 'check_return' in kwtypes and 'check_annot_only' in kwtypes)
    ):
        func = types[0]
        annotations = kwtypes.get('annotations', None)
        check_return = kwtypes.get('check_return', False)
        check_annot_only = kwtypes.get('check_annot_only', False)
        if isinstance(func, method): args = (func.__self__,) + args; func = func.__func__
        wrapped_func = func
        while hasattr(wrapped_func, '__wrapped__'): wrapped_func = wrapped_func.__wrapped__
        func_info = get_func_info(wrapped_func)
        f_name, _, _, f_varnames, *_ = func_info
        ret_annot = wrapped_func.__annotations__.get('return')
        if annotations is None: annotations = [wrapped_func.__annotations__.get(v, None) for v in f_varnames]
        def nest(a):
            if isinstance(a, str): return tag(a)
            if isinstance(a, tuple): return union(*a)
            if isinstance(a, list): return [nest(t) for t in a]
            if isinstance(a, dict): return {k: nest(t) for k, t in a.items()}
            if not isinstance(a, type) and isinstance(a, callable) or isinstance(a, tag('Type')): return instance_satisfies(a)
            return a
        annotations = [{k:nest(v) for k, v in a.items()} if isinstance(a, dict) else nest(a) for a in annotations]
        def wrapper(*args, **kwargs):
            # values = get_arg_values(func_info, *args, **kwargs)
            # check_annotations(func_info, values, annotations)
            match_and_check(func_info, annotations, *args, **kwargs)
            if check_annot_only: return lambda: None
            ret_value = func(*args, **kwargs)
            if check_return and ret_annot is not None:
                if not isinstance(ret_value, ret_annot):
                    raise TypeHintError(f"{f_name}() should have output of type {get_annot(a)}, but got {repr(ret_value)} of type {get_annot(ret_value.__class__)}")
            return ret_value
        return wrapper
    # Usage 3: @typehint(int, float, var1=str, var2=dict)
    else:
        @decorator
        def induced_decorator(func):
            __return__type__ = kwtypes.pop('__return__', None)
            annotations = get_arg_values(get_func_info(func), *types, **kwtypes)
            if __return__type__ is not None: annotations['return'] = __return__type__
            return typehint(func, annotations=annotations)
        return induced_decorator

if __name__ == "__main__":
    ...




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of different functions."
)

__all__ = """
    empty_function
    const_function
    identity_function
""".split()

def empty_function(*args, **kwargs):
    """ Just an empty function, one can use it as a placeholder. """
    pass

def const_function(a):
    """Create a function that always returns `a`. """
    def f(*args, **kwargs): return a
    return f

def identity_function(*x):
    """A funtion that returns the input, just as it was. """
    if len(x) == 1: return x[0]
    else: return x
